﻿using System;
using System.Web;
using System.Web.Configuration;
using System.Collections.Generic;
using System.Text;
using System.Text.RegularExpressions;
using System.Data;
using System.Data.OleDb;
using System.Globalization;
using VisualSite.Entities;

namespace VisualSite
{
    /// <summary>
    /// Data Access Layer containing
    /// all calls made to the database
    /// from the rest of the application
    /// </summary>
    public class DAL
    {

        #region Connection String

        /// <summary>
        /// Returns the connection string from web.config
        /// </summary>
        public static string ConnectionString
        {
            get { return WebConfigurationManager.ConnectionStrings[Constants.DEFAULT_ENVIRONMENT].ConnectionString; }
        }

        private static string _dataFile = null;
        /// <summary>
        /// Returns the file path to the Access database
        /// as a DataSource parameter in the web.config
        /// ConnectionStrings section
        /// </summary>
        public static string DataFile
        {
            get
            {
                if (String.IsNullOrEmpty(_dataFile))
                try
                {
                    if (!String.IsNullOrEmpty(ConnectionString))
                        using (OleDbConnection connection = new OleDbConnection(DAL.ConnectionString))
                            _dataFile = Regex.Replace(connection.DataSource, @"\|datadirectory\|", AppDomain.CurrentDomain.GetData("DataDirectory").ToString(), RegexOptions.IgnoreCase);
                }
                catch { }
                return _dataFile;
            }
        }

        #endregion

        #region Connection Test

        /// <summary>
        /// Returns true if we can read into the database
        /// </summary>
        /// <param name="ignoreException">Boolean to signal if we should ignore exceptions</param>
        /// <returns></returns>
        public static bool CheckRead(bool ignoreException)
        {
            try
            {
                ExecuteNonQuery("SELECT @@IDENTITY");
                return true;
            }
            catch (Exception ex) {
                if (!ignoreException)
                    throw ex; 
            }
            return false;
        }

        /// <summary>
        /// Returns true if we can write into the database
        /// </summary>
        /// <param name="ignoreException">Boolean to signal if we should ignore exceptions</param>
        /// <returns></returns>
        public static bool CheckWrite(bool ignoreException)
        {
            try
            {
                ExecuteNonQuery("CREATE TABLE TestConnection ([ID] INTEGER)");
                ExecuteNonQuery("DROP TABLE TestConnection");
                return true;
            }
            catch (Exception ex)
            {
                if (!ignoreException)
                    throw ex;
            }
            return false;
        }

        #endregion

        #region Page

        /// <summary>
        /// Gets a complete list of pages
        /// </summary>
        /// <returns></returns>
        public static List<Page> GetPages()
        {
            DataTable matches = ExecuteRowset("SELECT [ID], [Title], [Description], [RedirectUrl], [Body], [Author], [PublishDate], [ParentID], [Template], [Display], [DisplayOrder] FROM [Page] ORDER BY [ID]");
            return GetPages(matches);
        }

        /// <summary>
        /// Gets a list of pages filtered by ParentID
        /// </summary>
        /// <param name="parentID"></param>
        /// <returns></returns>
        public static List<Page> GetPagesByParentID(int parentID)
        {
            DataTable matches = ExecuteRowset(String.Format("SELECT [ID], [Title], [Description], [RedirectUrl], [Body], [Author], [PublishDate], [Template], [Display], [DisplayOrder] FROM [Page] WHERE [ParentID] = {0} ORDER BY [DisplayOrder]", parentID));
            return GetPages(matches);
        }

        /// <summary>
        /// Gets a list of menu items for a page
        /// </summary>
        /// <param name="currentPageID"></param>
        /// <returns></returns>
        public static List<Page> GetMenuItems(int currentPageID)
        {
            DataTable matches = ExecuteRowset(String.Format("SELECT [ID], [Title], [Display] FROM [Page] WHERE ([ID] <> {0}) AND (([ParentID] = 0) OR ([ParentID] = 1) OR ([ParentID] IN (SELECT [ID] FROM [Page] P WHERE ([ParentID] = 1)))) ORDER BY [ParentID], [DisplayOrder]", currentPageID));
            return GetPages(matches);
        }

        /// <summary>
        /// Gets a list of menu items excluding the current page
        /// but making sure that the parent of the current page is displayed
        /// in the list, this is used for populating the Parent dropdown
        /// </summary>
        /// <param name="sibblingID"></param>
        /// <returns></returns>
        public static List<Page> GetMenuItemsIncludingParent(int currentPageID)
        {
            int parentID = Utils.PageCatalog[currentPageID].ParentID;
            DataTable matches = ExecuteRowset(String.Format("SELECT [ID], [Title], [Display] FROM [Page] WHERE ([ID] <> {0}) AND (([ID] = {1}) OR ([ParentID] = 0) OR ([ParentID] = 1) OR ([ParentID] IN (SELECT [ID] FROM [Page] P WHERE ([ParentID] = 1)))) ORDER BY [ParentID], [DisplayOrder]", currentPageID, parentID));
            return GetPages(matches);
        }

        /// <summary>
        /// Gets a page from its ID
        /// </summary>
        /// <param name="id"></param>
        /// <returns></returns>
        public static Page GetPage(int id)
        {
            Page result = null;
            DataTable matches = ExecuteRowset(String.Format("SELECT [ID], [Title], [Description], [RedirectUrl], [Body], [Author], [PublishDate], [Template], [ParentID], [Display], [DisplayOrder] FROM [Page] WHERE [ID] = {0}", id));
            if (matches != null && matches.Rows.Count > 0)
            {
                result = GetPage(matches.Rows[0]);
            }
            return result;
        }

        /// <summary>
        /// Updates a page in the database
        /// </summary>
        /// <param name="id"></param>
        /// <param name="title"></param>
        /// <param name="description"></param>
        /// <param name="body"></param>
        /// <param name="parentID"></param>
        public static void UpdatePage(int id, string title, string description, string body, int parentID) 
        {
            ExecuteNonQuery(
                String.Format("UPDATE [Page] SET [Title] = \"{0}\", " + 
                                                "[Description] = \"{1}\", " +
                                                "[Body] = \"{2}\", " +
                                                "[Author] = \"{3}\", " +
                                                "[PublishDate] = \"{4}\", " +
                                                "[ParentID] = {5} " +
                                                "WHERE [ID] = {6}",
                                                Sanitise(title), Sanitise(description), Sanitise(body), Utils.CurrentUser, DateTime.Now.ToString(Constants.DATE_NEUTRAL_FORMAT), parentID, id)
                                                ); 
        }

        /// <summary>
        /// Updates a page in the database
        /// </summary>
        /// <param name="id"></param>
        /// <param name="title"></param>
        /// <param name="parentID"></param>
        /// <param name="redirectUrl"></param>
        /// <param name="display"></param>
        /// <param name="displayOrder"></param>
        public static void UpdatePage(int id, string title, int parentID, string redirectUrl, bool display, int displayOrder)
        {
            ExecuteNonQuery(
                String.Format("UPDATE [Page] SET [Title] = \"{0}\", " + 
                                                "[ParentID] = {1}, " +
                                                "[RedirectUrl] = \"{2}\", " +
                                                "[Display] = {3}, " +
                                                "[DisplayOrder] = {4} " +
                                                "WHERE [ID] = {5}",
                                                Sanitise(title), parentID, Sanitise(redirectUrl), display, displayOrder, id)
                                                ); 
        }

        /// <summary>
        /// Inserts a new page into the database
        /// </summary>
        /// <param name="title"></param>
        /// <param name="description"></param>
        /// <param name="body"></param>
        /// <param name="parentID"></param>
        /// <returns></returns>
        public static int InsertPage(string title, string description, string body, int parentID)
        {
            int result = -1;
            ExecuteNonQuery(
                String.Format("INSERT INTO [Page] ([Title],[Description],[Body],[Author],[PublishDate],[ParentID],[Display], [DisplayOrder]) " +
                                " VALUES (\"{0}\", \"{1}\", \"{2}\", \"{3}\", \"{4}\", {5}, TRUE, 1)",
                                Sanitise(title), Sanitise(description), Sanitise(body), Utils.CurrentUser, DateTime.Now.ToString(Constants.DATE_NEUTRAL_FORMAT), parentID)
                               );
            DataTable table = ExecuteRowset("SELECT MAX([ID]) FROM [Page]");
            if (table.Rows.Count > 0)
            {
                result = Convert.ToInt32(table.Rows[0][0]);
            }
            return result;
        }

        /// <summary>
        /// Deletes a page from the database
        /// </summary>
        /// <param name="id"></param>
        public static void DeletePage(int id)
        {
            // Delete great grandchildren
            ExecuteNonQuery(String.Format("DELETE FROM [Page] WHERE [ParentID] IN (SELECT [ID] FROM [Page] WHERE [ParentID] IN (SELECT [ID] FROM [Page] WHERE [ParentID] = {0}))", id));
            // Delete grandchildren
            ExecuteNonQuery(String.Format("DELETE FROM [Page] WHERE [ParentID] IN (SELECT [ID] FROM [Page] WHERE [ParentID] = {0})", id));
            // Delete children
            ExecuteNonQuery(String.Format("DELETE FROM [Page] WHERE [ParentID] = {0}", id));
            // Delete page
            ExecuteNonQuery(String.Format("DELETE FROM [Page] WHERE [ID] = {0}", id));
        }

        /// <summary>
        /// Takes a System.Data.DataRow and returns
        /// a Page object, used for data operations
        /// </summary>
        /// <param name="row"></param>
        /// <returns></returns>
        private static Page GetPage(DataRow row)
        {
            Page result = null;
            try
            {
                result = new Page(Convert.ToInt32(row["ID"]), row["Title"] as string);
                DataColumnCollection columns = row.Table.Columns;
                if (columns.Contains("Description"))
                    result.Description = row["Description"] as string;
                if (columns.Contains("RedirectUrl"))
                    result.RedirectUrl = row["RedirectUrl"] as string;
                if (columns.Contains("Body"))
                    result.Body = row["Body"] as string;
                if (columns.Contains("Author"))
                    result.Author = row["Author"] as string;
                if (columns.Contains("ParentID"))
                    result.ParentID = Convert.ToInt32(row["ParentID"]);
                if (columns.Contains("Template"))
                    result.Template = row["Template"] as string;
                if (columns.Contains("Display"))
                    result.Display = Convert.ToBoolean(row["Display"]);
                if (columns.Contains("DisplayOrder"))
                    result.DisplayOrder =Convert.ToInt32(row["DisplayOrder"]);
                // Publish date goes last in case there is an error parsing it
                if (columns.Contains("PublishDate"))
                    if (!String.IsNullOrEmpty(row["PublishDate"].ToString()))
                        result.PublishDate = DateTime.Parse(row["PublishDate"].ToString());
            }
            catch { }
            return result;
        }

        /// <summary>
        /// Gets a list of pages from a System.Data.DataTable input
        /// parameter.
        /// </summary>
        /// <param name="table"></param>
        /// <returns></returns>
        private static List<Page> GetPages(DataTable table)
        {
            List<Page> result = new List<Page>();
            foreach (DataRow row in table.Rows)
            {
                Page page = GetPage(row);
                if (page != null)
                    result.Add(page);
            }
            return result;
        }

        #endregion

        #region Setting

        /// <summary>
        /// Gets all the settings from the database
        /// </summary>
        /// <returns></returns>
        public static List<Setting> GetSettings()
        {
            List<Setting> result = new List<Setting>();
            DataTable matches = ExecuteRowset("SELECT [ID], [Name], [Value] FROM [Setting] ORDER BY [ID]");
            foreach (DataRow row in matches.Rows)
            {
                try
                {
                    result.Add(new Setting(Convert.ToInt32(row["ID"]), row["Name"] as string, row["Value"] as string));
                }
                catch { }
            }
            return result;
        }

        /// <summary>
        /// Updates a setting in the database filtering by id
        /// </summary>
        /// <param name="id"></param>
        /// <param name="value"></param>
        public static void UpdateSetting(int id, string value)
        {
            ExecuteNonQuery(String.Format("UPDATE [Setting] SET [Value] = \"{0}\" WHERE [ID] = {1}", Sanitise(value), id));
        }

        /// <summary>
        /// Updates a setting in the database filtering by name
        /// </summary>
        /// <param name="name"></param>
        /// <param name="value"></param>
        public static void UpdateSetting(string name, string value)
        {
            ExecuteNonQuery(String.Format("UPDATE [Setting] SET [Value] = \"{0}\" WHERE [Name] = \"{1}\"", Sanitise(value), Sanitise(name)));
        }

        #endregion

        #region User

        /// <summary>
        /// Returns a list of all users in the database
        /// </summary>
        /// <returns></returns>
        public static List<User> GetUsers()
        {
            List<User> result = new List<User>();
            DataTable matches = ExecuteRowset("SELECT [ID], [Username], [Password], [LockedDate] FROM [User] ORDER BY [ID]");
            foreach (DataRow row in matches.Rows)
            {
                DateTime lockedDate = DateTime.MinValue;
                try { lockedDate = DateTime.Parse(row["LockedDate"].ToString()); }
                catch { }
                try { result.Add(new User(Convert.ToInt32(row["ID"]), row["Username"] as string, row["Password"] as string, lockedDate)); }
                catch { }
            }
            return result;
        }

        /// <summary>
        /// Returns a user from the database filtering by username
        /// </summary>
        /// <param name="username"></param>
        /// <returns></returns>
        public static User GetUser(string username)
        {
            User result = null;
            DataTable matches = ExecuteRowset(String.Format("SELECT [ID], [Username], [Password], [LockedDate] FROM [User] WHERE [Username] = '{0}'", Sanitise(username)));
            if (matches != null && matches.Rows.Count > 0)
            {
                DateTime lockedDate = DateTime.MinValue;
                try { lockedDate = DateTime.Parse(matches.Rows[0]["LockedDate"].ToString()); }
                catch { }
                try { result = new User(Convert.ToInt32(matches.Rows[0]["ID"]), matches.Rows[0]["Username"] as string, matches.Rows[0]["Password"] as string, lockedDate); } 
                catch { }
            }
            return result;
        }

        /// <summary>
        /// Updates a user in the database
        /// </summary>
        /// <param name="id"></param>
        /// <param name="username"></param>
        /// <param name="password"></param>
        /// <param name="lockedDate"></param>
        public static void UpdateUser(int id, string username, string password, DateTime lockedDate)
        {
            ExecuteNonQuery(String.Format("UPDATE [User] SET [Username] = \"{0}\", [Password] = \"{1}\", [LockedDate] = \"{2}\" WHERE [ID] = {3}", Sanitise(username), Sanitise(password), lockedDate.ToString(Constants.DATE_NEUTRAL_FORMAT), id));
        }

        /// <summary>
        /// Inserts a user in the database
        /// </summary>
        /// <param name="username"></param>
        /// <param name="password"></param>
        public static void InsertUser(string username, string password)
        {
            ExecuteNonQuery(String.Format("INSERT INTO [User] ([Username],[Password], [LockedDate]) VALUES (\"{0}\", \"{1}\", NULL)", Sanitise(username), Sanitise(password)));
        }

        /// <summary>
        /// Deletes a user from the database
        /// </summary>
        /// <param name="id"></param>
        public static void DeleteUser(int id)
        {
            if (id > 1)
                ExecuteNonQuery(String.Format("DELETE FROM [User] WHERE [ID] = {0}", id));
            else
                throw new Exception("Sorry, can't delete the admin user");
        }
        #endregion

        #region Schema Commands 

        /// <summary>
        /// Refreshes the database by recreating
        /// all relevant tables with a basic
        /// set of data, including the username
        /// and password for later administration
        /// </summary>
        /// <param name="sitename">Name of the Site</param>
        /// <param name="username">Username for the CMS</param>
        /// <param name="password">Password for the CMS</param>
        public static void RefreshDatabase(string sitename, string username, string password)
        {
            // Delete tables
            try { ExecuteNonQuery("DROP TABLE [Setting]"); }
            catch { }
            try { ExecuteNonQuery("DROP TABLE [Page]"); }
            catch { }
            try { ExecuteNonQuery("DROP TABLE [User]"); }
            catch { }
            // Create Tables
            ExecuteNonQuery(@"CREATE TABLE [Setting] (
                                [ID] COUNTER CONSTRAINT PK_Setting PRIMARY KEY,
                                [Name] NVARCHAR(50) WITH COMPRESSION NOT NULL,
                                [Value] MEMO WITH COMPRESSION)
                                ");
            ExecuteNonQuery(@"CREATE TABLE [User] (
                                [ID] COUNTER CONSTRAINT PK_User PRIMARY KEY,
                                [Username] NVARCHAR(50) WITH COMPRESSION NOT NULL,
                                [Password] NVARCHAR(50) WITH COMPRESSION NOT NULL,
                                [LockedDate] DATETIME)
                                ");
            ExecuteNonQuery(@"CREATE TABLE [Page] (
                                [ID] COUNTER CONSTRAINT PK_User PRIMARY KEY,
                                [ParentID] INTEGER NOT NULL,
                                [Title] NVARCHAR(50) WITH COMPRESSION NOT NULL,
                                [Description] NVARCHAR(50) WITH COMPRESSION,
                                [RedirectUrl] NVARCHAR(250) WITH COMPRESSION,
                                [Body] MEMO WITH COMPRESSION,
                                [Author] NVARCHAR(50) WITH COMPRESSION,
                                [PublishDate] DATETIME,
                                [Template] NVARCHAR(50) WITH COMPRESSION,
                                [Display] BIT,
                                [DisplayOrder] INTEGER)
                                ");
            // Fill in Settings
            string description = (Utils.Locale.StartsWith("es")) ? "Funciona!" : "It Works!";
            ExecuteNonQuery(@"INSERT INTO [Setting] ([Name], [Value])
                                VALUES ('SiteDescription', '" + description  + "')");
            ExecuteNonQuery(@"INSERT INTO [Setting] ([Name], [Value])
                                VALUES ('Keywords', '" + String.Join(", ", sitename.Split(' ')) + "')");
            ExecuteNonQuery(@"INSERT INTO [Setting] ([Name], [Value])
                                VALUES ('SiteTemplate', NULL)");
            ExecuteNonQuery(@"INSERT INTO [Setting] ([Name], [Value])
                                VALUES ('FriendlyUrls', 'OFF')");
            ExecuteNonQuery(@"INSERT INTO [Setting] ([Name], [Value])
                                VALUES ('Compression', 'ON')");
            ExecuteNonQuery(@"INSERT INTO [Setting] ([Name], [Value])
                                VALUES ('Backup', 'OFF')");
            ExecuteNonQuery(@"INSERT INTO [Setting] ([Name], [Value])
                                VALUES ('CacheDuration', '30')");
            ExecuteNonQuery(@"INSERT INTO [Setting] ([Name], [Value])
                                VALUES ('Version', '" + Constants.APP_VERSION + "')");
            ExecuteNonQuery(@"INSERT INTO [Setting] ([Name], [Value])
                                VALUES ('Locale', '" + Utils.Locale + "')");
            ExecuteNonQuery(@"INSERT INTO [Setting] ([Name], [Value])
                                VALUES ('InstallState', 'Installing..')");

            // Fill in Home
            string body = (Utils.Locale.StartsWith("es")) ? "Mi primera página" : "My first page";
            string now = DateTime.Now.ToString(Constants.DATE_NEUTRAL_FORMAT);
            ExecuteNonQuery(@"INSERT INTO [Page] ([ParentID], [Title], [Description], [RedirectUrl], [Body], [Author], [PublishDate], [Template], [Display], [DisplayOrder])
                                VALUES ('0', '" + sitename + "', NULL, NULL, '" + body + "', NULL, '" + now + "', NULL, TRUE, 1)");
            // Fill in User
            ExecuteNonQuery(@"INSERT INTO [User] ([Username], [Password], [LockedDate])
                                VALUES ('" + Sanitise(username) + "', '" + Sanitise(password) + "', NULL)");
        }
        #endregion

        #region Data Commands

        /// <summary>
        /// Private method which executes SQL code on the database
        /// and returns a set of results. Note a try/catch block
        /// is used to provent locking up the database when a
        /// query fails or the connection drops out.
        /// </summary>
        /// <param name="commandText">SQL Query</param>
        /// <returns></returns>
        private static DataTable ExecuteRowset(string commandText)
        {
            DataTable result = new DataTable(String.Empty);
            if (!String.IsNullOrEmpty(commandText))
            {
                using (OleDbConnection connection = new OleDbConnection(DAL.ConnectionString))
                {
                    OleDbCommand command = connection.CreateCommand();
                    command.CommandType = CommandType.Text;
                    command.CommandText = commandText;
                    try
                    {
                        // Open connection and execute SQL command
                        connection.Open();
                        OleDbDataReader reader = command.ExecuteReader();
                        if (reader.HasRows)
                        {
                            result.BeginLoadData();
                            result.Load(reader);
                            result.EndLoadData();
                        }
                        reader.Close();
                    }
                    catch (Exception ex)
                    {
                        throw ex;
                    }
                    finally
                    {
                        command.Dispose();
                        connection.Close();
                        connection.Dispose();
                    }
                }
            }
            return result;
        }

        /// <summary>
        /// Executes a SQL command without
        /// returning any results, this is used
        /// for INSERT, UPDATE or DELETE operations
        /// </summary>
        /// <param name="commandText">SQL Query</param>
        private static void ExecuteNonQuery(string commandText)
        {
            try
            {
                ExecuteRowset(commandText);
            }
            catch (Exception ex)
            {
                throw ex;
            }
        }

        /// <summary>
        /// Removes damaging input to avoid SQL injection
        /// </summary>
        /// <param name="input"></param>
        /// <returns></returns>
        private static string Sanitise(string input) 
        {
            // Sanitise input to avoid SQL injection
            if (input == null)
                return input;
            else 
                return input.Replace("\"", "\"\"");
        }
        #endregion
    }
}
