﻿using System;
using System.IO;
using System.IO.Compression;
using System.Collections.Generic;
using System.Web;
using System.Threading;
using System.Globalization;

namespace VisualSite
{
    /// <summary>
    /// This module performs various tasks needed at the beginning
    /// of each request.
    /// 
    /// 1. Redirects user to the "/Install" directory
    ///    unless installation has completed
    /// 
    /// 2. Adds current locale to current Thread for 
    ///    successful localisation
    /// 
    /// 3. Handles URL rewriting for "Friendly Url" logic
    /// 
    /// 4. Handles web page compression to optimise
    ///    bandwidth usage
    /// 
    /// Add to web.config as follows:
    /// 
    /// 	<httpModules>
    ///      <add name="VisualSiteHttpModule" type="VisualSite.HttpModule" />
    ///     </httpModules>
    /// </summary>
    public class HttpModule : System.Web.IHttpModule
    {
        // Set some fields we are going to need
        HttpApplication _application = null;
        string _installDir = null;
        string _redirectUrl = null;

        public void Init(HttpApplication application)
        {
            // Add our event handler to the queue
            application.PreRequestHandlerExecute += new System.EventHandler(Application_PreRequestHandlerExecute);
            application.BeginRequest += new System.EventHandler(Application_BeginRequest);
            // Set local field for access in the method below
            _application = application;
        }

        public void Dispose()
        {
        }

        /// <summary>
        /// EventHandler method to attach to the Application.BeginRequest event
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Application_BeginRequest(object sender, System.EventArgs e)
        {
            // Load the current request path into a local variable
            string requestPath = _application.Request.Path;
            bool isAspx = requestPath.ToLower().EndsWith(".aspx");

            // *****************************************************************
            // This part handles redirect when first installing the Application
            // *****************************************************************

            // Redirects user to the "/Install" directory
            // unless installation is complete.

            // Ignore other file extensions, focus only on redirecting pages
            if (isAspx)
            {
                // Check install state. Once ready we should stop redirecting requests
                if (!Utils.IsInstallationComplete)
                {
                    // Initialise values if they are null
                    if (String.IsNullOrEmpty(_installDir))
                        _installDir = Path.GetFullPath(_application.Server.MapPath(Constants.INSTALL_FOLDER)).TrimEnd(Path.DirectorySeparatorChar);

                    if (String.IsNullOrEmpty(_redirectUrl))
                        // Work out url to install folder
                        _redirectUrl = _application.Request.ApplicationPath.TrimEnd('/') +
                            _installDir.Replace(_application.Request.PhysicalApplicationPath, "/");

                    // Check Installation directory, if it exists then redirect the user there
                    FileInfo currentFile = new FileInfo(_application.Server.MapPath(requestPath));
                    if (Directory.Exists(_installDir) && (currentFile.DirectoryName != _installDir))
                    {
                        _application.Response.Redirect(_redirectUrl, false);
                    }
                }
            }

            // *****************************************************************
            // This part ensures localization works properly
            // *****************************************************************

            // Enforce locale on current thread
            // this is needed because each request runs on a new thread 
            if (isAspx)
                try
                {
                    CultureInfo culture = new CultureInfo(Utils.Locale);
                    if (Thread.CurrentThread.CurrentUICulture.Name != culture.Name) Thread.CurrentThread.CurrentUICulture = culture;
                    if (Thread.CurrentThread.CurrentCulture.Name != culture.Name) Thread.CurrentThread.CurrentCulture = culture;
                }
                catch { }

            // *****************************************************************
            // This part handles URL rewriting
            // *****************************************************************

            // Rewrite URL Logic
            // Use the folder id to identify page object
            //
            // Convert: ~/12.page-name.aspx
            // To: ~/Default.aspx?id=12
            //

            if (isAspx)
            {
                string[] request = requestPath.Substring(requestPath.LastIndexOf('/') + 1).Split('.');
                if (request.Length == 2)
                {
                    string title = request[0];
                    try
                    {
                        // Getting the catalog will fail at installation (no database)
                        // ignore UrlRewriting at this stage
                        Dictionary<int, VisualSite.Entities.Page> pageCatalog = Utils.PageCatalog;
                        if (Utils.UrlCatalogByUrl.ContainsKey(title))
                            _application.Context.RewritePath(String.Format(VisualSite.Constants.RAW_URL, Utils.UrlCatalogByUrl[title]));
                    }
                    catch { }
                }
            }
        }

        /// <summary>
        /// EventHandler method to attach to the Application.PreRequestHandlerExecute event
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Application_PreRequestHandlerExecute(object sender, EventArgs e)
        {
            // *****************************************************************
            // This part handles web page compression to reduce loading time
            // *****************************************************************

            // Checks the Accept-Encoding header for "gzip" or "deflate" formats
            // Encodes HTTP output accordingly

            if (Utils.UseCompression)
            {
                // Set local variables for headers and un-compressed output
                string acceptEncoding = _application.Request.Headers["Accept-Encoding"];
                Stream prevUncompressedStream = _application.Response.Filter;

                // Check handler is a valid web page
                if (!(_application.Context.CurrentHandler is System.Web.UI.Page ||
                    _application.Context.CurrentHandler.GetType().Name == "SyncSessionlessHandler") ||
                    _application.Request["HTTP_X_MICROSOFTAJAX"] != null)
                    return;

                if (acceptEncoding == null || acceptEncoding.Length == 0)
                    return;

                acceptEncoding = acceptEncoding.ToLower();

                // Use "deflate" if available
                if (acceptEncoding.Contains("deflate") || acceptEncoding == "*")
                {
                    _application.Response.Filter = new DeflateStream(prevUncompressedStream,
                        CompressionMode.Compress);
                    _application.Response.AppendHeader("Content-Encoding", "deflate");
                }
                // Check for "gzip" format if not
                else if (acceptEncoding.Contains("gzip"))
                {
                    _application.Response.Filter = new GZipStream(prevUncompressedStream,
                        CompressionMode.Compress);
                    _application.Response.AppendHeader("Content-Encoding", "gzip");
                }
            }
        }
    }
}
