﻿using System;
using System.Collections.Generic;
using System.Web;
using System.Text.RegularExpressions;
using VisualSite;
using VisualSite.Entities;

/// <summary>
/// SearchManager is a class which obtains performance-obtimised search
/// queries for site content. It uses regular expressions and the cached
/// PageCatalog hashtable to improve search efficiency.
/// </summary>
public class SearchManager
{
    /// <summary>
    /// Gets search results for a particular query.
    /// </summary>
    /// <param name="query"></param>
    /// <returns></returns>
    public List<SearchResult> GetResults(string query)
    {
        // Initialise resultset
        List<SearchResult> result = new List<SearchResult>();
        // Start search
        if (!String.IsNullOrEmpty(query))
        {
            // Reduce query size (avoid DOS attacks)
            if (query.Length > 100) query = query.Substring(0, 100);
            // Searches ignore punctuation (,.!?") and markup (<b>hello</b>)
            query = Utils.RemoveNonAlphaNumeric(Utils.RemoveMarkup(query));
            // Regex gives us the ability to search word limits with the "\b" pattern
            string queryPattern = @"\b" + query + @"\b";
            // Loop through all entries in page catalog
            foreach (KeyValuePair<int, Page> entry in Utils.PageCatalog)
            {
                // Get a page from the catalog
                Page page = entry.Value;
                // Dont search hidden pages
                if (page.Display)
                {
                    // Check the title for a match (ignore case)
                    if (Regex.Match(page.Title, queryPattern, RegexOptions.IgnoreCase).Success)
                        result.Add(GetSearchResult(page, query));
                    // Check the description (skip if empty)
                    else if (!String.IsNullOrEmpty(page.Description)
                        && Regex.Match(page.Description, queryPattern, RegexOptions.IgnoreCase).Success)
                        result.Add(GetSearchResult(page, query));
                    // Last of all, since its the heaviest, check the body (but skip if empty)
                    else if (!String.IsNullOrEmpty(page.Body)
                        && Regex.Match(page.Body, queryPattern, RegexOptions.IgnoreCase).Success)
                        result.Add(GetSearchResult(page, query));
                }
            }
        }
        // Order results by relevance (number of matches)
        if (result.Count > 0)
            result.Sort(delegate(SearchResult a, SearchResult b) { return b.Matches.CompareTo(a.Matches); });

        // Return results
        return result;
    }

    /// <summary>
    /// Returns a search result object per page.
    /// Search entities contain several bit of information
    /// that are search-specific like number of matches etc.
    /// </summary>
    /// <param name="page"></param>
    /// <param name="query"></param>
    /// <returns></returns>
    private SearchResult GetSearchResult(Page page, string query)
    {
        // Initialise
        int matches = 0;
        string blurb = "...";
        string bodyText = Utils.RemoveMarkup(page.Body as string);
        string queryPattern = @"\b" + query + @"\b";
        // Obtain match count in title and description (skip the latter if its empty)
        if (Regex.Match(page.Title, queryPattern, RegexOptions.IgnoreCase).Success) matches++;
        if (!String.IsNullOrEmpty(page.Description) 
            && Regex.Match(page.Description, queryPattern, RegexOptions.IgnoreCase).Success) matches++;
        // Before proceeding, check body contents
        if (!String.IsNullOrEmpty(bodyText))
        {
            // Count number of matches
            Match match = Regex.Match(bodyText, queryPattern, RegexOptions.IgnoreCase);
            while (match.Success)
            {
                matches++;
                match = match.NextMatch();
            }
            // Load matches again, then create a blurb
            // depending on location of the match and size of body text
            match = Regex.Match(bodyText, queryPattern, RegexOptions.IgnoreCase);
            if (match.Success)
            {
                if (bodyText.Length < 100)
                    blurb = Regex.Replace(bodyText, queryPattern, "<b>$&</b>", RegexOptions.IgnoreCase);
                else
                {
                    // Creating this blurb is a bit messy but there is no better way I can think of
                    string formatString = "{0}";
                    int start = match.Index - 25;
                    if (start < 0) start = 0;
                    else formatString = "... " + formatString;
                    int length = query.Trim().Length + 100;
                    if (start + length > bodyText.Length) length = bodyText.Length - start;
                    else formatString = formatString + " ...";
                    blurb = String.Format(formatString, Regex.Replace(bodyText.Substring(start, length), queryPattern, "<b>$&</b>", RegexOptions.IgnoreCase));
                }
            }
            else
            {
                // If we cant find a match then show the first 100 characters
                if (bodyText.Length < 100)
                    blurb = bodyText;
                else
                    blurb = String.Format("{0} ...", bodyText.Substring(0, 100));
            }
        }
        // Return the search result
        return new SearchResult(page.ID, page.Title, Utils.GetBreadcrumb(page.ID), page.Description, blurb, Utils.GetUrl(page.ID), matches);
    }
}
