﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Data;
using System.Security.Permissions;
using System.Web;
using System.Web.Configuration;
using System.Text.RegularExpressions;
using VisualSite.Entities;

namespace VisualSite
{
    /// <summary>
    /// An extremely simple SiteMapProvider 
    /// Generates an ASP.NET compatible sitemap based on 
    /// database calls returning a hierarchical set of pages
    /// 
    /// Add to web.config as follows:
    /// 
    ///   <siteMap defaultProvider="SimpleSiteMapProvider">
    ///    <providers>
    ///      <add name="SimpleSiteMapProvider" type="VisualSite.SimpleSiteMapProvider" />
    ///    </providers>
    ///   </siteMap>
    ///   
    /// </summary>
    [AspNetHostingPermission(SecurityAction.Demand, Level = AspNetHostingPermissionLevel.Minimal)]
    public class SimpleSiteMapProvider : StaticSiteMapProvider
    {
        // Some basic state to help track the initialization state of the provider.
        private bool initialized = false;
        public virtual bool IsInitialized
        {
            get
            {
                return initialized;
            }
        }
        
        /// <summary>
        /// Returns the root node of the SiteMap
        /// </summary>
        public override SiteMapNode RootNode
        {
            get
            {
                SiteMapNode temp = null;
                try
                {
                    temp = BuildSiteMap();
                }
                catch { }
                return temp;
            }
        }

        /// <summary>
        /// Gets the root node
        /// </summary>
        /// <returns></returns>
        protected override SiteMapNode GetRootNodeCore()
        {
            return RootNode;
        }

        /// <summary>
        /// Override the base class Initialise. This is used to initialise
        /// any properties to pass through from the web.config
        /// </summary>
        /// <param name="name"></param>
        /// <param name="attributes"></param>
        public override void Initialize(string name, NameValueCollection attributes)
        {
            if (IsInitialized)
                return;

            base.Initialize(name, attributes);

            initialized = true;
        }

        /// <summary>
        /// Overrides the base class Clear method to clean up
        /// any collections or other state that an instance of this may hold.
        /// </summary>
        protected override void Clear()
        {
            lock (this)
            {
                HttpContext.Current.Cache.Remove("rootNode");
                base.Clear();
            }
        }

        /// <summary>
        /// Use this method externally to clear the 
        /// sitemap from HTTP cache. We want to do this in case
        /// the structure changes suddenly by adding new pages
        /// or modifying the location of existing ones.
        /// Once called again the sitemap will rebuild a 
        /// new version of itself and add it to cache automatically.
        /// </summary>
        public static void ClearFromCache()
        {
            HttpContext.Current.Cache.Remove("rootNode");
        }

        /// <summary>
        /// Build an in-memory representation from persistent
        /// storage (database in this case), and return the 
        /// root node of the site map.
        /// </summary>
        /// <returns></returns>
        public override SiteMapNode BuildSiteMap()
        {
            SiteMapNode rootNode;

            // Since the SiteMap class is static, make sure that it is
            // not modified while the site map is built.
            lock (this)
            {

                // If there is no initialization, this method is being
                // called out of order.
                if (!IsInitialized)
                {
                    throw new Exception("BuildSiteMap called incorrectly.");
                }

                // Check the cache for a sitemap
                rootNode = HttpContext.Current.Cache["rootNode"] as SiteMapNode;

                // If there is no root node, then there is no site map.
                if (null == rootNode)
                {
                    // Start with a clean slate
                    Clear();

                    // Select the root node of the site map from your data source.
                    int id = -1;
                    Page page = DAL.GetPage(1);

                    // Plug all the values into a node
                    if (page != null)
                    {
                        id = page.ID;
                        string title = page.Title;
                        string url = Utils.GetUrl(id);
                        NameValueCollection siteMapAttributes = new NameValueCollection();
                        siteMapAttributes.Add("id", id.ToString());
                        siteMapAttributes.Add("display", page.Display.ToString());
                        siteMapAttributes.Add("redirect", page.RedirectUrl);
                        // Create a SiteMapNode that references the current StaticSiteMapProvider.
                        rootNode = new SiteMapNode(this,
                                                     id.ToString(),
                                                     url,
                                                     title,
                                                     page.Description,
                                                     null,
                                                     siteMapAttributes,
                                                     null,
                                                     null);

                    }
                    else return null;

                    // Add the node as the root
                    AddNode(rootNode);
                    // Add all the children to it
                    AddChildren(rootNode, id);
                    // This is where we use some caching magic to improve performance
                    // Its expensive to rebuild the sitemap with every HTTP request
                    // so we stick it in the cache until expiry, so we can 
                    // later make a call to that instead.
                    Utils.AddToCache("rootNode", rootNode);
                }
                return rootNode;
            }
        }

        /// <summary>
        /// Recursive method which calls itself to 
        /// add all remaining nodes of the site map
        /// </summary>
        /// <param name="parentNode"></param>
        /// <param name="parentID"></param>
        private void AddChildren(SiteMapNode parentNode, int parentID)
        {
            // Get child pages
            List<Page> pages = DAL.GetPagesByParentID(parentID);
            // Add to parent
            foreach (Page page in pages)
            {
                SiteMapNode childNode = null;
                NameValueCollection siteMapAttributes = new NameValueCollection();
                int id = page.ID;
                string title = page.Title;
                siteMapAttributes.Add("id", id.ToString());
                siteMapAttributes.Add("display", page.Display.ToString());
                siteMapAttributes.Add("redirect", page.RedirectUrl);
                string url = Utils.GetUrl(id);
                // Check Redirect
                if(!String.IsNullOrEmpty(page.RedirectUrl)) {
                    // SiteMap Cant handle absolute urls
                    if (!System.Uri.IsWellFormedUriString(page.RedirectUrl, UriKind.Absolute))
                    {
                        System.Web.UI.Page handler = HttpContext.Current.Handler as System.Web.UI.Page;
                        if (handler != null)
                            url = handler.ResolveUrl(page.RedirectUrl);
                    }
                }
                childNode = new SiteMapNode(this,
                                             id.ToString(),
                                             url,
                                             title,
                                             page.Description,
                                             null,
                                             siteMapAttributes,
                                             new NameValueCollection(),
                                             null);
                // Use recursion to add child nodes.
                AddChildren(childNode, id);
                // Then use the base SiteMapNode AddNode method to add
                // the SiteMapNode to the ChildNodes collection.
                AddNode(childNode, parentNode);
            }
        }
    }
}