﻿using System;
using System.Data;
using System.Collections;
using System.Collections.Generic;
using System.Configuration;
using System.IO;
using System.Text.RegularExpressions;
using System.Security.AccessControl;
using System.Web;
using System.Web.Configuration;
using System.Web.Security;
using System.Web.UI;
using System.Web.UI.WebControls;
using System.Globalization;
using System.Threading;
using System.Runtime.InteropServices;
using VisualSite.Entities;

namespace VisualSite
{
    /// <summary>
    /// Contains a set of static method calls to use thorughout the VisualSite application. 
    /// These have all been placed in a single class to avoid the possible confusion of users
    /// trying to figure out an over-engineered object hierarchy. 
    /// </summary>
    public class Utils
    {

        #region State Information

        /// <summary>
        /// Returns true if user is authenticated
        /// </summary>
        public static bool IsUserAuthenticated
        {
            get { return !String.IsNullOrEmpty(CurrentUser); }
        }

        /// <summary>
        /// Name of the current user
        /// </summary>
        public static string CurrentUser
        {
            get { return HttpContext.Current.User.Identity.Name as string; }
        }

        /// <summary>
        /// Returns true if use is the admin user
        /// </summary>
        public static bool UserIsAdmin
        {
            get
            {
                if (!IsUserAuthenticated) return false;
                try { return DAL.GetUser(CurrentUser).ID == 1; }
                catch { }
                return false;
            }
        }

        /// <summary>
        /// Catalog of all Pages in the site
        /// </summary>
        public static Dictionary<int, VisualSite.Entities.Page> PageCatalog
        {
            get
            {
                Dictionary<int, VisualSite.Entities.Page> result = HttpContext.Current.Cache["PageCatalog"] as Dictionary<int, VisualSite.Entities.Page>;
                if (result == null)
                {
                    result = new Dictionary<int, VisualSite.Entities.Page>();
                    foreach (VisualSite.Entities.Page page in DAL.GetPages())
                    {
                        result.Add(page.ID, page);
                    }
                    Utils.AddToCache("PageCatalog", result);
                }
                return result;
            }
        }

        /// <summary>
        /// ID of the current page
        /// </summary>
        public static int PageID
        {
            get
            {
                int result = 0;
                // Try to obtain current page ID from request string 
                // (its less processor-heavy than consulting SiteMap.CurrentNode)
                try { result = Convert.ToInt32(HttpContext.Current.Request.QueryString[Constants.ID_KEY]); }
                catch { }
                // If that fails, consult the sitemap
                if (result == 0)
                    try { result = Convert.ToInt32(SiteMap.CurrentNode["id"]); }
                    catch { }
                return result;
            }
        }

        /// <summary>
        /// Title of the current page
        /// </summary>
        public static string PageTitle
        {
            get
            {
                // Obtain the Page Title from catalog
                string result = String.Empty;
                VisualSite.Entities.Page currentPage = null;
                if (PageID > 0)
                {
                    try
                    {
                        currentPage = PageCatalog[PageID];
                        result = currentPage.Title;
                    }
                    catch { }
                }
                else if (PageID < 0) result = Localise("navNewPage");
                if (String.IsNullOrEmpty(result))
                {
                    // If that fails use file name instead
                    // "/MyFolder/File-Manager.aspx" will return "File Manager"
                    // "/Admin/Default.aspx" will return "Admin"
                    try
                    {
                        string requestedPath = HttpContext.Current.Server.MapPath(HttpContext.Current.Request.Path);
                        FileInfo file = new FileInfo(requestedPath);
                        string pathName = String.Empty;
                        if (file != null)
                            if (file.Name.ToLower().EndsWith("default.aspx"))
                                pathName = file.Directory.Name;
                            else
                                pathName = file.Name.Replace(file.Extension, String.Empty);
                        // Try to localise the path name
                        if (String.IsNullOrEmpty(Localise("nav" + pathName)))
                            result = pathName;
                        else
                            result = Localise("nav" + pathName);
                    }
                    catch { }
                }
                return result;
            }
        }

        /// <summary>
        /// Depth of the current page in terms of how many parents
        /// separate it from root node
        /// </summary>
        public static int CurrentDepth
        {
            get
            {
                /// Find out dept of current page
                /// Use a copy of the catalog to improve performance
                /// Admin = -1
                /// -------
                /// Home = 0
                /// |- Sub = 1
                /// |- Sub = 1
                ///    |- SubSub = 2
                ///    
                int result = -1;
                SiteMapNode currentNode = SiteMap.CurrentNode;
                Dictionary<int, VisualSite.Entities.Page> pageCatalog = PageCatalog;
                if (currentNode != null && pageCatalog.Count > 0)
                {
                    result++;
                    int nodeID = Convert.ToInt32(currentNode["id"]);
                    while (nodeID != 1)
                    {
                        result++;
                        nodeID = pageCatalog[nodeID].ParentID;
                    }
                }
                return result;
            }
        }

        /// <summary>
        /// The Url of the website
        /// </summary>
        public static string SiteUrl
        {
            get
            {
                // TODO: This could be cached.. or maybe there is another method call?
                return HttpContext.Current.Request.Url.OriginalString.Replace(HttpContext.Current.Request.Url.PathAndQuery, "");
            }
        }

        /// <summary>
        /// The title of the website
        /// Derived from the title of the homepage
        /// Use caching to improve performance
        /// </summary>
        public static string SiteTitle
        {
            get
            {
                string result = HttpContext.Current.Cache.Get("SiteTitle") as string;
                if (String.IsNullOrEmpty(result))
                {
                    result = PageCatalog[1].Title;
                    if (String.IsNullOrEmpty(result))
                        try
                        {
                            result = DAL.GetPage(1).Title;
                        }
                        catch { }
                    AddToCache("SiteTitle", result);
                }
                return result;
            }
            set
            {
                VisualSite.Entities.Page page = DAL.GetPage(1);
                DAL.UpdatePage(page.ID, value, page.Description, page.Body, page.ParentID);
                HttpContext.Current.Cache.Remove("SiteTitle");
                AddToCache("SiteTitle", value);
            }
        }

        /// <summary>
        /// A method which returns a breadcrumb string for a given page
        /// </summary>
        /// <param name="pageID">ID of the page from which the breadcrumb is rendered</param>
        /// <returns></returns>
        public static string GetBreadcrumb(int pageID)
        {
            string result = "";
            List<VisualSite.Entities.Page> breadcrumb = GetBreadcrumbList(pageID);
            for (int i = 0; i < breadcrumb.Count; i++)
            {
                if (i < breadcrumb.Count - 1)
                    result += breadcrumb[i].Title + " &raquo; ";
                else
                    result += breadcrumb[i].Title;
            }
            return result;
        }

        /// <summary>
        /// A method which returns a list of pages
        /// ending with node from which the breadcrumb is rendered.
        /// </summary>
        /// <param name="pageID">ID of the page from which the breadcrumb is rendered</param>
        /// <returns></returns>
        public static List<VisualSite.Entities.Page> GetBreadcrumbList(int pageID)
        {
            List<VisualSite.Entities.Page> pageList = new List<VisualSite.Entities.Page>();
            Dictionary<int, VisualSite.Entities.Page> pageCatalog = PageCatalog;
            AddBreadcrumbAncestor(pageID, ref pageList, ref pageCatalog);
            return pageList;
        }

        /// <summary>
        /// Private recursion method which takes in a list of pages
        /// to add to as we navigate up a hierarchy of nodes
        /// in the page catalog. Note: Referencing is used to improve 
        /// performance since copying parameters is more expensive.
        /// </summary>
        /// <param name="pageID"></param>
        /// <param name="pages"></param>
        private static void AddBreadcrumbAncestor(int pageID, ref List<VisualSite.Entities.Page> pages, ref Dictionary<int, VisualSite.Entities.Page> catalog)
        {
            if (catalog.ContainsKey(pageID))
            {
                AddBreadcrumbAncestor(catalog[pageID].ParentID, ref pages, ref catalog);
                pages.Add(catalog[pageID]);
            }
        }
        #endregion

        #region Redirection Tools

        /// <summary>
        /// Catalog matching Urls to page ids
        /// Used for urlRewriting
        /// </summary>
        public static Dictionary<string, int> UrlCatalogByUrl
        {
            get
            {
                Dictionary<string, int> result = HttpContext.Current.Cache["UrlCatalogByUrl"] as Dictionary<string, int>;
                if (result == null)
                {
                    result = new Dictionary<string, int>();
                    foreach (VisualSite.Entities.Page page in PageCatalog.Values)
                    {
                        // Remove punctuation and replace spaces with dashes
                        string title = Regex.Replace(page.Title, @"[^A-Za-z0-9]", "-");
                        // Check if there is another page using the same name
                        while (result.ContainsKey(title) || File.Exists(HttpContext.Current.Server.MapPath(String.Format(Constants.FRIENDLY_URL, title))))
                            title = title + "-";
                        // Add to catalog
                        result.Add(title, page.ID);
                    }
                    AddToCache("UrlCatalogByUrl", result);
                }
                return result;
            }
        }

        /// <summary>
        /// Catalog matching ids to urls 
        /// (backward reference of UrlCatalogByUrl)
        /// Used for urlRewriting
        /// </summary>
        public static Dictionary<int, string> UrlCatalogByID
        {
            get
            {
                Dictionary<int, string> result = HttpContext.Current.Cache["UrlCatalogByID"] as Dictionary<int, string>;
                if (result == null)
                {
                    result = new Dictionary<int, string>();
                    foreach (KeyValuePair<string, int> entry in UrlCatalogByUrl)
                        result.Add(entry.Value, entry.Key);
                    AddToCache("UrlCatalogByID", result);
                }
                return result;
            }
        }

        /// <summary>
        /// Returns a root relative path ~/filename.aspx
        /// for a given page ID
        /// </summary>
        /// <param name="id"></param>
        /// <returns></returns>
        public static string GetUrl(int id)
        {
            string url = String.Format(Constants.RAW_URL, id);
            // Check if we are using Friendly Urls and catalog contains this page
            if (Utils.GetSetting("FriendlyUrls") == "ON" && UrlCatalogByID.ContainsKey(id))
                // Replace non-alphanumeric characters with dashes 
                url = String.Format(Constants.FRIENDLY_URL, UrlCatalogByID[id]);
                //url = String.Format(Constants.FRIENDLY_URL, id, Regex.Replace(PageCatalog[id].Title, @"[^\w]", "-"));
            return url;
        }

        /// <summary>
        /// Redirects the user to the homepage
        /// </summary>
        public static void GoHome()
        {
            Go(1);
        }

        /// <summary>
        /// Redirects the user to a page by using
        /// a specific ID
        /// </summary>
        /// <param name="pageID"></param>
        public static void Go(int pageID)
        {
            HttpContext context = HttpContext.Current;
            System.Web.UI.Page page = context.Handler as System.Web.UI.Page;
            if (PageCatalog.ContainsKey(pageID))
            {
                if (String.IsNullOrEmpty(PageCatalog[pageID].RedirectUrl))
                    context.Response.Redirect(page.ResolveUrl(GetUrl(pageID)), true);
                else
                    context.Response.Redirect(page.ResolveUrl(PageCatalog[pageID].RedirectUrl), true);
            }
        }

        #endregion

        #region Language Tools

        /// <summary>
        /// Returns current application locale by checking
        /// current settings and making any necessary adjustments
        /// to the culture of the Current Thread for string operations
        /// </summary>
        public static string Locale
        {
            get
            {
                // Look for locale in cache
                string result = HttpContext.Current.Cache["Locale"] as string;
                // If empty search the database
                if (String.IsNullOrEmpty(result))
                    try
                    {
                        result = Utils.GetSetting("Locale");
                        AddToCache("Locale", result);
                    }
                    catch { }
                // If still empty use default value
                if (String.IsNullOrEmpty(result))
                    result = Constants.DEFAULT_LOCALE;
                return result;
            }
            set
            {
                // Current languages allowed are English and Spanish
                if (value != null)
                    if (value.StartsWith("en") || value.StartsWith("es"))
                    {
                        // Add to cache and database
                        AddToCache("Locale", value);
                        try { DAL.UpdateSetting("Locale", value); }
                        catch { }
                        // Set thread locale
                        try
                        {
                            CultureInfo culture = new CultureInfo(value);
                            Thread.CurrentThread.CurrentUICulture = culture;
                            Thread.CurrentThread.CurrentCulture = culture;
                        }
                        catch { }
                    }
            }
        }

        /// <summary>
        /// Localises a value using the GlobalResourceObject
        /// </summary>
        /// <param name="key"></param>
        /// <returns></returns>
        public static string Localise(string key)
        {
            // Use Global Resource Object to find localised string
            return HttpContext.GetGlobalResourceObject("Resources", key) as string; ;
        }

        #endregion

        #region Settings

        /// <summary>
        /// Returns a particular application setting
        /// </summary>
        /// <param name="key">Name of the setting</param>
        /// <returns>Value of the setting or NULL if there is no match</returns>
        public static string GetSetting(string key)
        {
            // Look for setting in cache
            string result = HttpContext.Current.Cache.Get(key) as string;
            // Cant find it? Load from database and add to cache.
            if (result == null)
            {
                // Initialise settings table
                Hashtable settingTable = new Hashtable();
                // Load results from database
                List<Setting> settings = new List<Setting>();
                // Get settings from DB. Ignore errors
                try { settings = DAL.GetSettings(); }
                catch { }
                // Proceed only if we managed to retrieve data successfully
                if (settings.Count > 0)
                {
                    // Add each setting to output hash
                    foreach (Setting setting in settings)
                        if (!settingTable.Contains(setting.Name))
                            settingTable.Add(setting.Name, setting.Value);
                    // Store everything in cache
                    foreach (string setting in settingTable.Keys)
                        AddToCache(setting, (settingTable[setting] == null) ? String.Empty : settingTable[setting]);
                }
                // Load data
                result = settingTable[key] as string;
            }
            return result;
        }

        /// <summary>
        /// Returns the template file from settings
        /// checking to see if the file exists in the process
        /// </summary>
        public static string TemplateFile
        {
            get
            {
                string result = String.Empty;

                string siteTemplate = GetSetting("SiteTemplate");
                if (!String.IsNullOrEmpty(siteTemplate))
                    try
                    {
                        string siteTemplatePath = HttpContext.Current.Server.MapPath(siteTemplate);
                        if (System.IO.File.Exists(siteTemplatePath))
                            result = siteTemplate;
                    }
                    catch { }

                return result;
            }
        }

        /// <summary>
        /// Returns true if installation has completed
        /// </summary>
        public static bool IsInstallationComplete
        {
            get
            {
                return GetSetting("InstallState") == "Ready";
            }
        }

        /// <summary>
        /// Returns true if we are using Http compression
        /// </summary>
        public static bool UseCompression
        {
            get
            {
                return GetSetting("Compression") == "ON";
            }
        }

        #endregion

        #region Cache Management

        /// <summary>
        /// Adds an object to application cache for a timespan determined in 
        /// the database "CacheDuration" setting.
        /// </summary>
        /// <param name="key">Name of the cache parameter</param>
        /// <param name="value">Object to store in cache</param>
        public static void AddToCache(string key, object value)
        {
            double cacheDuration = 20.0;
            if (!String.IsNullOrEmpty(HttpContext.Current.Cache["CacheDuration"] as string))
                cacheDuration = Convert.ToDouble(HttpContext.Current.Cache["CacheDuration"]);
            HttpContext.Current.Cache.Remove(key);
            HttpContext.Current.Cache.Insert(key, value, null, DateTime.Now.AddMinutes(cacheDuration), TimeSpan.Zero);
        }

        /// <summary>
        /// Adds an object to application cache for a set number of minutes
        /// This is used for locking repeat attempts at logging in
        /// since the number of minutes will be different to the database setting.
        /// </summary>
        /// <param name="key">Name of the cache parameter</param>
        /// <param name="value">Object to store in cache</param>
        /// <param name="minutes">Number of minutes to hold in cache</param>
        public static void AddToCache(string key, object value, int minutes)
        {
            double cacheDuration = Convert.ToDouble(minutes);
            HttpContext.Current.Cache.Remove(key);
            HttpContext.Current.Cache.Insert(key, value, null, DateTime.Now.AddMinutes(cacheDuration), TimeSpan.Zero);
        }

        /// <summary>
        /// Clears application cache
        /// </summary>
        /// <returns></returns>
        public static string ClearCache()
        {
            string message = "";
            IDictionaryEnumerator entries = HttpContext.Current.Cache.GetEnumerator();
            while (entries.MoveNext())
            {
                HttpContext.Current.Cache.Remove(entries.Entry.Key as string);
                message += entries.Entry.Key as string + ", ";
            }
            return message;
        }

        #endregion

        #region Cookie Management

        /// <summary>
        /// Gets a Cookie value from the Users Browser
        /// </summary>
        /// <param name="key">Cookie key</param>
        /// <returns>Cookie value</returns>
        public static string GetCookie(string key)
        {
            string result = null;
            HttpCookie cookie = HttpContext.Current.Request.Cookies[key];
            if (cookie != null) result = cookie[key];
            return result;
        }

        /// <summary>
        /// Sets a Cookie value on the Users Browser
        /// </summary>
        /// <param name="key">Cookie key</param>
        /// <param name="value">Cookie value</param>
        public static void SetCookie(string key, string value)
        {
            HttpCookie cookie = new HttpCookie(key);
            cookie[key] = value;
            cookie.Expires = DateTime.Now.AddYears(1);
            HttpContext.Current.Response.Cookies.Add(cookie);
        }

        #endregion

        #region Text Formatting

        // Normalisation function to match text characters
        // Lowers case and adds space on each side (match exact words)
        // "Hello!" becomes " hello! "
        // "Hello there, Mary!" becomes " hello there, mary! "
        public static string LowerAndSpaceUp(string text)
        {
            if (!String.IsNullOrEmpty(text))
                text = " " + text.ToLower() + " ";
            return text;
        }

        // Normalisation function to match text characters
        // Removes punctuation (,.!?") 
        // "Hello!" becomes "hello"
        // "Hello there, Mary!" becomes "hello there mary"
        public static string RemoveNonAlphaNumeric(string text)
        {
            if (!String.IsNullOrEmpty(text))
                text = Regex.Replace(Regex.Replace(text, @"[^\w]", " "), @"\s+|\r|\n", " ").Trim();
            return text;
        }

        // Normalisation function to match text characters
        // Removes XML/HTML Markup
        // "<a href=""#">Click Here</a>" becomes "Click Here"
        // "Hello there, <b>Mary</b>!" becomes "Hello there, Mary!"
        public static string RemoveMarkup(string htmlcode)
        {
            if (!String.IsNullOrEmpty(htmlcode))
                return Regex.Replace(htmlcode, "</?[^>]*?>", "");
            return htmlcode;
        }

        #endregion

        #region FileSystem Operations

        /// <summary>
        /// Checks to see if database has already
        /// been backed up today
        /// </summary>
        public static bool IsDatabaseBackedUp
        {
            get
            {
                bool result = false;
                try
                {
                    result = File.Exists(DAL.DataFile.Replace(".mdb", "." + DateTime.Now.ToString("yyyy-MM-dd") + ".mdb.bak"));
                }
                catch { }
                return result;
            }
        }

        /// <summary>
        /// Backs up the database
        /// </summary>
        public static void BackupDatabase()
        {
            if (File.Exists(DAL.DataFile))
            {
                // Backup File
                string filePath = DAL.DataFile;
                string backupFilePath = filePath.Replace(".mdb", "." + DateTime.Now.ToString("yyyy-MM-dd") + ".mdb.bak");
                FileInfo file = new FileInfo(filePath);
                FileInfo backupFile = new FileInfo(backupFilePath);
                // Backups should only be made once a day max
                if (!backupFile.Exists)
                    file.CopyTo(backupFilePath);
            }
        }

        /// <summary>
        /// Private class encapsulating Windows OS
        /// functionality used for impersonation
        /// in the method further down
        /// </summary>
        private static class WinTools
        {
            internal const int LOGON32_LOGON_INTERACTIVE = 2;
            internal const int LOGON32_LOGON_NETWORK = 3;
            internal const int LOGON32_LOGON_BATCH = 4;
            internal const int LOGON32_LOGON_SERVICE = 5;
            internal const int LOGON32_LOGON_UNLOCK = 7;
            internal const int LOGON32_LOGON_NETWORK_CLEARTEXT = 8;
            internal const int LOGON32_LOGON_NEW_CREDENTIALS = 9;
            internal const int LOGON32_PROVIDER_DEFAULT = 0;

            [DllImport("advapi32.dll", SetLastError = true)]
            internal static extern int LogonUser(
                string lpszUsername,
                string lpszDomain,
                string lpszPassword,
                int dwLogonType,
                int dwLogonProvider,
                out IntPtr phToken
                );
            [DllImport("advapi32.dll", SetLastError = true)]
            internal static extern int ImpersonateLoggedOnUser(
                IntPtr hToken
            );

            [DllImport("advapi32.dll", SetLastError = true)]
            internal static extern int RevertToSelf();

            [DllImport("kernel32.dll", SetLastError = true)]
            internal static extern int CloseHandle(IntPtr hObject);
        }

        /// <summary>
        /// Sets permissions in a directory within the filesystem
        /// so that IIS can modify (read/write) files in it.
        /// This is done using a current Windows username and
        /// password combination.
        /// </summary>
        /// <param name="username"></param>
        /// <param name="password"></param>
        /// <param name="directory"></param>
        public static void SetModifyPermissions(string username, string password, string directory)
        {
            // Exit if input is not valid (password can be empty)
            if (String.IsNullOrEmpty(username) || String.IsNullOrEmpty(directory))
                return;

            if (!Directory.Exists(directory))
                return;

            // If there is no domain then use the current computer
            string domain = ".";
            if (username.Contains("\\"))
            {
                domain = username.Substring(0, username.IndexOf("\\") - 1);
                username = username.Substring(username.IndexOf("\\") + 1);
            }

            // Obtain the user IIS is running under
            string IISUserName = System.Security.Principal.WindowsIdentity.GetCurrent().Name;

            // Check to see if user/password combination can log in ok.
            IntPtr lnToken;
            int TResult = WinTools.LogonUser(username, domain, password,
                            WinTools.LOGON32_LOGON_NETWORK, WinTools.LOGON32_PROVIDER_DEFAULT,
                            out lnToken);

            // Exit method if user is not valid
            if (TResult < 1)
                return;

            try
            {
                // Impersonate user
                WinTools.ImpersonateLoggedOnUser(lnToken);
                // Execute code
                DirectoryInfo dirInfo = new DirectoryInfo(directory);
                DirectorySecurity dirSecurity = dirInfo.GetAccessControl();
                FileSystemAccessRule modifyRole = new FileSystemAccessRule(IISUserName, FileSystemRights.Modify | FileSystemRights.DeleteSubdirectoriesAndFiles, InheritanceFlags.ObjectInherit | InheritanceFlags.ContainerInherit, PropagationFlags.None, AccessControlType.Allow);
                dirSecurity.SetAccessRule(modifyRole);
                dirInfo.SetAccessControl(dirSecurity);
            }
            catch { }
            // Revert impersonation
            WinTools.RevertToSelf();
            WinTools.CloseHandle(lnToken);

        }

        #endregion

    }
}