﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Web;
using System.Web.UI;
using System.Web.UI.WebControls;
using System.Reflection;
using System.Data.OleDb;
using VisualSite;
using VisualSite.Entities;

public partial class Install_Default : System.Web.UI.Page
{
    // Initialise some values
    protected int Step = 1;
    protected int Pages = 0;
    protected int Users = 0;
    protected bool DataOk = false;
    protected int Images = 0;
    protected bool FileWriteOk = false;
    protected string DBVersion = "N/A";
    protected string Language = "N/A";
    protected string CurrentUser = System.Security.Principal.WindowsIdentity.GetCurrent().Name;

    protected void Page_Init(object sender, EventArgs e)
    {
        // After Installation do not allow random access to 
        // the installation folder unless the admin user is logged in
        if (Utils.IsInstallationComplete && !Utils.UserIsAdmin)
            Utils.GoHome();
        // Redirect to Step 1 if accessed directly
        else if (String.IsNullOrEmpty(Request.QueryString["s"]))
            Response.Redirect(Request.Path + "?s=1");
        else 
        {
            Step = 1;
            try
            {
                Step = Convert.ToInt32(Request.QueryString["s"]);
            }
            catch { }
            if (Step == 4) CheckData();
            if (Step == 5) CheckFileWrite();
            if (Step == 6)
            {
                // Signal ready
                DAL.UpdateSetting("InstallState", "Ready");
                Utils.ClearCache();
            }
        }
    }

    /// <summary>
    /// Checks the database exists and passes all basic
    /// tests to verify format and key data
    /// </summary>
    protected void CheckData() {
        if (DAL.CheckRead(true) && DAL.CheckWrite(true)) {
            try
            {
                // Check data is there
                Pages = DAL.GetPages().Count;
                Users = DAL.GetUsers().Count;
                // Set data for UI
                foreach (Setting setting in DAL.GetSettings())
                {
                    if (setting.Name == "Locale")
                        Language = setting.Value;
                    if (setting.Name == "Version")
                        DBVersion = setting.Value;
                }
                // Verify users and DB version
                if (DAL.GetPage(1) == null)
                    throw new Exception(); // Home page missing, you need a homepage
                if (Users < 1) 
                    throw new Exception(); // There needs to be at least 1 User in the VisualSite database
                if (Constants.APP_VERSION != DBVersion)
                    throw new Exception(); // Using an older version of the database. 
                // Confirm data
                DataOk = true;
            }
            catch { }
        }
    }

    /// <summary>
    /// Checks the file folder can be written to
    /// and how many files are stored in there
    /// </summary>
    protected void CheckFileWrite()
    {
        DirectoryInfo fileDir = new DirectoryInfo(Server.MapPath(Constants.FILE_FOLDER));
        if (fileDir.Exists)
        {
            try
            {
                fileDir.CreateSubdirectory("Test");
                fileDir.GetDirectories("Test")[0].Delete();
                FileWriteOk = true;
                Images = fileDir.GetFiles().Length;
            }
            catch { }
        }
    }

    /// <summary>
    /// Generates a new Access (2000) database file 
    /// using the Jet provider. (Step 2)
    /// </summary>
    /// <param name="sender"></param>
    /// <param name="e"></param>
    protected void GenerateDatabase(object sender, EventArgs e)
    {
        if (!String.IsNullOrEmpty(DAL.DataFile) && DAL.DataFile.ToLower().EndsWith(".mdb"))
        {
            // Delete file if it already exists
            if (File.Exists(DAL.DataFile))
                try
                {
                    File.Delete(DAL.DataFile);
                }
                catch { }

            // Brute force database file creation. 
            // Note: this will not work unless IIS
            // is permissioned to write files
            try
            {
                // Use a late bound COM object to create a new catalog. 
                // This is so we avoid an interop assembly. 
                Type catType = Type.GetTypeFromProgID("ADOX.Catalog");
                string connString = "Provider=Microsoft.Jet.OLEDB.4.0;Data Source=" + DAL.DataFile;
                object o = Activator.CreateInstance(catType);
                catType.InvokeMember("Create", BindingFlags.InvokeMethod, null, o, new object[] { connString });
                using (OleDbConnection cnn = new OleDbConnection(connString))
                {
                    cnn.Open();
                    OleDbCommand cmd = cnn.CreateCommand();
                    cmd.CommandText = "SELECT @@IDENTITY";
                    cmd.ExecuteNonQuery();
                }
            }
            catch { }
        }
    }

    /// <summary>
    /// Permissions the database folder to allow
    /// modify by IIS user. (Step 2)
    /// Generates a new Access (2000) database file 
    /// using the Jet provider. 
    /// </summary>
    /// <param name="sender"></param>
    /// <param name="e"></param>
    protected void GenerateDatabaseUsingCredentials(object sender, EventArgs e)
    {
        if (String.IsNullOrEmpty(txtWinUsername.Text) || String.IsNullOrEmpty(txtWinPassword.Text))
            return;
        FileInfo dbFile = new FileInfo(VisualSite.DAL.DataFile);
        if (dbFile.Directory.Exists)
        {
            // Fix permissions in step 2
            Utils.SetModifyPermissions(txtWinUsername.Text, txtWinPassword.Text, dbFile.Directory.FullName);
            GenerateDatabase(null, null);
        }
        // Fix permissions in step 5 while we are there
        DirectoryInfo fileFolder = new DirectoryInfo(Server.MapPath(Constants.FILE_FOLDER));
        if (fileFolder.Exists)
            Utils.SetModifyPermissions(txtWinUsername.Text, txtWinPassword.Text, fileFolder.FullName);
    }

    /// <summary>
    /// Permissions the file folder to allow
    /// IIS to read and write to it. (Step 5)
    /// </summary>
    /// <param name="sender"></param>
    /// <param name="e"></param>
    protected void FixFilePermissionsUsingCredentials(object sender, EventArgs e)
    {
        DirectoryInfo fileFolder = new DirectoryInfo(Server.MapPath(Constants.FILE_FOLDER));
        if (String.IsNullOrEmpty(txtWinUsername2.Text) || String.IsNullOrEmpty(txtWinPassword2.Text))
            return;
        if (fileFolder.Exists)
            Utils.SetModifyPermissions(txtWinUsername2.Text, txtWinPassword2.Text, fileFolder.FullName);
    }

    /// <summary>
    /// Refreshes database with new tables and data
    /// </summary>
    /// <param name="sender"></param>
    /// <param name="e"></param>
    protected void RefreshDatabase(object sender, EventArgs e)
    {
        if (DAL.CheckRead(true) && DAL.CheckWrite(true)) {
            if (!String.IsNullOrEmpty(txtSiteName.Text) && !String.IsNullOrEmpty(txtUsername.Text))
            try
            {
                // Backup database
                if (!File.Exists(DAL.DataFile + ".bak"))
                    File.Copy(DAL.DataFile, DAL.DataFile + ".bak");
                // Refresh database
                DAL.RefreshDatabase(txtSiteName.Text, txtUsername.Text, txtPassword.Text);
                // Clear cache and refresh
                Utils.ClearCache();
                Response.Redirect(Request.Url.OriginalString, true);
            }
            catch { }
        }
    }

    /// <summary>
    /// Delete all files in File repository and refresh the page
    /// </summary>
    /// <param name="sender"></param>
    /// <param name="e"></param>
    protected void DeleteFiles(object sender, EventArgs e)
    {
        DirectoryInfo fileDir = new DirectoryInfo(Server.MapPath(Constants.FILE_FOLDER));
        if (fileDir.Exists)
        {
            foreach (FileInfo file in fileDir.GetFiles())
                file.Delete();
            Response.Redirect(Request.Url.OriginalString, true);
        }
    }
}
